import numpy as np


def get_affine_transform(moving, fixed, with_ones=False):
    """
    :param moving: point cloud 3 x N
    :param fixed: point cloud 3 x N
    :param with_ones: False
    :return: 4 x 4 affine matrix
    """
    if (with_ones):
        pass
    else:
        extra_one_row = np.ones((1, moving.shape[1]))
        moving = np.vstack((moving, extra_one_row))
        fixed = np.vstack((fixed, extra_one_row))
    return np.matmul(fixed, np.linalg.pinv(moving))


# http://www.sci.utah.edu/~shireen/pdfs/tutorials/Elhabian_ICP09.pdf
def get_similar_transform(moving, fixed):
    """
    :param moving: point cloud 3 x N
    :param fixed: point cloud 3 x N
    :return: s (1) , R (3 x 3) , t (3 x 1)
    """
    com_target = np.mean(fixed, 1, keepdims=True)  # 3 x 1
    com_source = np.mean(moving, 1, keepdims=True)  # 3 x 1

    # eliminate translation
    Yprime = fixed[:3, :] - com_target[:3, :]
    Pprime = moving[:3, :] - com_source[:3, :]

    # use quarternions
    Px = Pprime[0, :]
    Py = Pprime[1, :]
    Pz = Pprime[2, :]

    Yx = Yprime[0, :]
    Yy = Yprime[1, :]
    Yz = Yprime[2, :]

    Sxx = np.sum(Yx * Px)
    Sxy = np.sum(Px * Yy)
    Sxz = np.sum(Px * Yz)

    Syx = np.sum(Py * Yx)
    Syy = np.sum(Py * Yy)
    Syz = np.sum(Py * Yz)

    Szx = np.sum(Pz * Yx)
    Szy = np.sum(Pz * Yy)
    Szz = np.sum(Pz * Yz)

    Nmatrix = [[Sxx + Syy + Szz, Syz - Szy, -Sxz + Szx, Sxy - Syx],
               [-Szy + Syz, Sxx - Szz - Syy, Sxy + Syx, Sxz + Szx],
               [Szx - Sxz, Syx + Sxy, Syy - Szz - Sxx, Syz + Szy],
               [-Syx + Sxy, Szx + Sxz, Szy + Syz, Szz - Syy - Sxx]]

    [V, D] = np.linalg.eig(Nmatrix)  # TODO values, vectors
    # optimal quarternion vector corresponds to the largest eigen value
    idx = V.argsort()[::-1]
    V = V[idx]
    D = D[:, idx]

    q = D[0]  # first value is largest value!
    q0 = q[0]
    q1 = q[1]
    q2 = q[2]
    q3 = q[3]

    Qbar = [[q0, -q1, -q2, -q3],
            [q1, q0, q3, -q2],
            [q2, -q3, q0, q1],
            [q3, q2, -q1, q0]]

    Q = [[q0, -q1, -q2, -q3],
         [q1, q0, -q3, q2],
         [q2, q3, q0, -q1],
         [q3, -q2, q1, q0]]

    R = np.matmul(np.transpose(Qbar), Q)
    R = R[1:, 1:]

    # compute scaling
    Sp = 0
    D = 0

    for i in range(Yprime.shape[1]):
        D += np.matmul(np.transpose(Yprime[:, i]), Yprime[:, i])
        Sp += np.matmul(np.transpose(Pprime[:, i]), Pprime[:, i])

    s = np.sqrt(D / Sp)
    t = com_target[:3, :] - s * np.matmul(R, com_source[:3, :])
    A = np.zeros((4, 4))
    A[:3, :3] = s * R
    A[:3, 3:4] = t
    A[3, 3] = 1
    return A



import numpy as np
from platymatch.detect_nuclei.ss_log import sphere_intersection, find_spheres


def test_sphere_intersection_1():
    r1 = np.random.randint(100)
    r2 = np.random.randint(100)
    d = r1 + r2 + np.random.randint(100)
    assert sphere_intersection(r1, r2, d) == 0


def test_sphere_intersection_2():
    r1 = np.random.randint(100)
    r2 = np.random.randint(100)
    d = 0
    if r1 < r2:
        assert sphere_intersection(r1, r2, d) == 4 / 3 * np.pi * (r1 ** 3)
    else:
        assert sphere_intersection(r1, r2, d) == 4 / 3 * np.pi * (r2 ** 3)


def test_sphere_intersection_3():
    r1 = np.random.randint(100)
    r2 = r1/2
    d = r2/2
    
    assert sphere_intersection(r1, r2, d) == 4 / 3 * np.pi * (r2 ** 3)


def test_sphere_intersection_4():
    r1 = np.random.randint(100)
    r2 = np.random.randint(100)
    d = r1+r2

    assert sphere_intersection(r1, r2, d) == 0.0

def test_find_spheres_1():
    im = np.zeros((100, 100, 100))
    x1, y1, z1 = np.random.randint(20, 40, 3)  # sphere 1 between 10:40, 10:40, 10:40
    r1 = 20

    x2, y2, z2 = np.random.randint(60, 90, 3)
    r2 = 20

    for z_i in range(z1 - r1, z1 + r1 + 1):
        for y_i in range(y1 - r1, y1 + r1 + 1):
            for x_i in range(x1 - r1, x1 + r1 + 1):
                d = np.linalg.norm((x_i - x1) ** 2 + (y_i - y1) ** 2 + (z_i - z1) ** 2)
                if (d <= r1):
                    im[z_i, y_i, x_i] = 100 - d

    for z_i in range(z2 - r2, z2 + r2 + 1):
        for y_i in range(y2 - r2, y2 + r2 + 1):
            for x_i in range(x2 - r2, x2 + r2 + 1):
                d = np.linalg.norm((x_i - x2) ** 2 + (y_i - y2) ** 2 + (z_i - z2) ** 2)
                if (d <= r2):
                    im[z_i, y_i, x_i] = 100 - d

    _, peaks_subset, _, _, _ = find_spheres(im, scales=range(1, 10), anisotropy_factor=1.0)
    assert len(peaks_subset) == 2



import numpy as np
import pandas as pd
from qtpy.QtWidgets import QFileDialog

def _visualize_nuclei(self, nuclei):
    for layer in self.viewer.layers:
        if layer.name == 'points-' + self.viewer.layers[self.images_combo_box.currentIndex()].name:
            self.viewer.layers.remove('points-' + self.viewer.layers[self.images_combo_box.currentIndex()].name)

    point_properties = {
        'color': nuclei[:, 1] + nuclei[:, 2] + nuclei[:, 3],  # use z coordinate as placeholder for color!
        'radius': np.sqrt(3)*nuclei[:, 0]
    }

    self.viewer.add_points(nuclei[:, 1:], size=np.sqrt(3) * 2 * nuclei[:, 0], properties=point_properties,
                           face_color='color', face_colormap='viridis', opacity=0.5,
                           name="points-" + self.viewer.layers[self.images_combo_box.currentIndex()].name)

def _browse_detections(self):
    name = QFileDialog.getOpenFileName(self, 'Open File')  # this returns a tuple!
    print("Opening detections {} ******".format(name[0]))
    if self.header_checkbox.isChecked():
        detections_df = pd.read_csv(name[0], skiprows=[0], delimiter=' ', header=None)
    else:
        detections_df = pd.read_csv(name[0], skiprows=None, delimiter=' ', header=None)
    detections_numpy = detections_df.to_numpy()
    ids = detections_numpy[:, 0]  # first column should contain ids
    detections = detections_numpy[:, 1:4] # N x 3
    if self.izyx_checkbox.isChecked():
        pass
    else:
        detections = np.flip(detections, 1) # N x 3

    return detections.transpose(), ids.transpose() # 3 x N, 1 x N


def _browse_transform(self):
    name = QFileDialog.getOpenFileName(self, 'Open File')  # this returns a tuple!
    print("Opening transform {} ******".format(name[0]))
    transform_df = pd.read_csv(name[0], skiprows=None, delimiter=' ', header= None)
    transform_numpy = transform_df.to_numpy()
    assert (transform_numpy.shape==(4,4)), 'Loaded transform does not hasve shape 4 x 4'
    return transform_numpy # 4 x 4




def get_centroid(detections, transposed = True):
    """
    :param detections: N x 3/4 or 3/4 x N
    :return: 3 x 1
    """
    if (transposed):  # N x 4
        return np.mean(detections[:, :3], 0, keepdims=True)
    else: # 4 x N
        return np.mean(detections[:3, :], 1, keepdims=True)

def get_mean_distance(detections, transposed = False):
    """
    :param detections: 3 x N
    :return:
    """

    if (transposed): # 3 x N
        pass
    else: # N x 3
        detections = detections.transpose() # N x 3

    if detections.shape[1] == 4:
        detections = detections[:, :3] # N x 3
    pair_wise_distance = []
    for i in range(detections.shape[0]):
        for j in range(i + 1, detections.shape[0]):
            pair_wise_distance.append(np.linalg.norm(detections[i] - detections[j]))
    return np.average(pair_wise_distance)

def get_error(moving_landmarks, fixed_landmarks):
    """

    :param moving_landmarks: 3 x N
    :param fixed_landmarks: 3 x N
    :return:
    """
    if (moving_landmarks is not None or fixed_landmarks is not None):
        residual = moving_landmarks - fixed_landmarks
        return np.mean(np.linalg.norm(residual, axis=0)) # axis= 0 means norm is taken along axis = 0
    else:
        return None


import numpy as np
from scipy.spatial import distance_matrix
from platymatch.estimate_transform.find_transform import get_affine_transform
from platymatch.estimate_transform.apply_transform import apply_affine_transform
from platymatch.utils.utils import get_error

def perform_icp(moving, fixed, icp_iterations = 50, transform='Affine'):
    if moving.shape[0] == 4:  # 4 x N
        moving = moving[:3, :] # 3 x N
    if fixed.shape[0] == 4:  # 4 x N
        fixed = fixed[:3, :] # 3 x N

    A_icp = np.identity(4)
    for i in range(icp_iterations):
        cost_matrix=distance_matrix(moving.transpose(), fixed.transpose()) # (N x 3, N x 3)
        i2=np.argmin(cost_matrix, 1)
        if(transform=='Affine'):
            A_est=get_affine_transform(moving, fixed[:, i2])
        elif(transform=='similar'):
            A_est = get_similar_transform(moving, fixed[:, i2])
        elif(transform=='rigid'):
            A_est = get_rigid_transform(moving, fixed[:, i2])
        moving = apply_affine_transform(moving, A_est)
        print("Residual at iteration {} is {}".format(str(i), get_error(moving, fixed[:, i2])))
        A_icp=np.matmul(A_est, A_icp)
    return A_icp


import numpy as np
from sklearn.decomposition import PCA
from platymatch.estimate_transform.find_transform import get_affine_transform, get_similar_transform
from platymatch.estimate_transform.apply_transform import *

def get_Y(z, x):
    y = np.cross(z, x)
    return y / np.linalg.norm(y)

def get_shape_context(neighbors, mean_dist, r_inner=1 / 8, r_outer=2, n_rbins=5, n_thetabins=6, n_phibins=12):
    """
    :param neighbors:  N-1 x 3
    :param mean_dist: 1
    :param r_inner:
    :param r_outer:
    :param n_rbins:
    :param n_thetabins:
    :param n_phibins:
    :return:
    """
    r = []
    theta = []
    phi = []
    r_edges = np.logspace(np.log10(r_inner), np.log10(r_outer), n_rbins)
    for neighbor in neighbors:
        x_ = neighbor[0]
        y_ = neighbor[1]
        z_ = neighbor[2]
        r_ = np.linalg.norm(neighbor)
        r.append(r_ / mean_dist)
        theta.append(np.arccos(z_ / r_))
        if (np.arctan2(y_, x_) < 0):
            phi.append(2 * np.pi + np.arctan2(y_, x_))
        else:
            phi.append(np.arctan2(y_, x_))
    index = get_bin_index(r, theta, phi, r_edges, n_rbins, n_thetabins, n_phibins)

    sc = np.zeros((n_rbins * n_thetabins * n_phibins))
    for i in range(n_rbins * n_thetabins * n_phibins):
        sc[i] = index.count(i)
    sc= sc/sc.sum()
    return sc



def get_bin_index(r, theta, phi, r_edges, n_rbins, n_thetabins, n_phibins):
    binIndex = []  # 0 ... 359 (12 * 5 * 6)
    for i in range(len(r)):  # i: index over all neighbors
        r_index = n_rbins - 1  # by default, it is assumed to be the last most ring

        theta_index = theta[i] // (np.pi / n_thetabins)
        phi_index = phi[i] // (2 * np.pi / n_phibins)
        for ind, edge in enumerate(r_edges):  # ind: over all possible edges of bins
            if (r[i] < edge):
                r_index = ind
                break;
        binIndex.append(r_index * n_thetabins * n_phibins + theta_index * n_phibins + phi_index)
    return binIndex


def transform(detection, x_vector, y_vector, z_vector, neighbors):
    """

    :param detection: 1 x 3
    :param x_vector: 1 x 3
    :param y_vector: 1 x 3
    :param z_vector: 1 x 3
    :param neighbors: (N-1) x 3
    :return:
    """
    x_1_3d = detection + x_vector
    y_1_3d = detection + y_vector
    z_1_3d = detection + z_vector
    A = np.ones((4, 4))
    A[0, :3] = detection
    A[1, :3] = x_1_3d
    A[2, :3] = y_1_3d
    A[3, :3] = z_1_3d
    A = np.transpose(A)
    B = np.array([[0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1], [1, 1, 1, 1]])
    T = np.matmul(B, np.linalg.inv(A))
    neighbors = np.hstack((neighbors, np.ones((neighbors.shape[0], 1))))
    points_transformed = np.matmul(T, np.transpose(neighbors))
    return np.transpose(points_transformed)[:, :3]



def get_unary_distance(sc1, sc2):
    """

    :param sc1: (360, )
    :param sc2: (360, )
    :return:
    """
    dist = 0
    for i in range(sc1.shape[0]):
        if (sc1[i] != sc2[i]):
            dist = dist + ((sc1[i] - sc2[i]) ** 2) / (sc1[i] + sc2[i])
    return 0.5 * dist



def do_ransac(moving_all, fixed_all, min_samples=4, trials=500, error=5, transform ='Affine'):
    """

    :param moving_all: 4 x N
    :param fixed_all: 4 x N
    :param min_samples:
    :param trials:
    :param error:
    :param transform:
    :return:
    """

    if moving_all.shape[0] == 4 or fixed_all.shape[0]==4:  # 4 x N
        moving_all = moving_all[:3, :] # 3 x N
        fixed_all = fixed_all[:3, :] # 3 x N

    inliers_best = 0
    A_best = np.ones((4, 4))
    for i in range(trials):
        indices = np.random.choice(fixed_all.shape[1], min_samples, replace=False)
        target_chosen = fixed_all[:, indices] # 3 x min_samples
        source_chosen = moving_all[:, indices] # 3 x min_samples

        if transform =='Affine':
            transform_matrix = get_affine_transform(source_chosen, target_chosen)
        elif transform =='Similar':
            transform_matrix = get_similar_transform(source_chosen, target_chosen)
        predicted_all = apply_affine_transform(moving_all, transform_matrix)  # 3 x N
        inliers = 0
        for index in range(fixed_all.shape[1]):
            d = np.linalg.norm(fixed_all[:, index] - predicted_all[:, index])
            if (d <= error):
                inliers += 1
        if (inliers > inliers_best):
            inliers_best = inliers
            A_best = transform_matrix
    return A_best, inliers_best




def get_unary(centroid, mean_distance, detections, transposed = False):
    """
    :param centroid: N x 3 or 3 x N (transposed = False)
    :param mean_distance:
    :param detections: N x 4
    :return:
    """
    if(transposed): # N x 3
        pass
    else: # 3 x N
        detections = detections.transpose()
        centroid = centroid.transpose()
    if detections.shape[1] == 4:  # N x 4
        detections = detections[:, :3] # N x 3
    sc = []
    sc2 = []
    pca = PCA(n_components=3)
    pca.fit(detections)
    V = pca.components_
    x0_vector = V[0:1, :] # 1 x 3
    x0_vector2 = -1 * x0_vector
    for queried_index, detection in enumerate(detections):
        neighbors = np.delete(detections, queried_index, 0) # (N-1) x 3
        z_vector = (detection - centroid) / np.linalg.norm(detection - centroid) # vector 1 --> Z
        x_vector = x0_vector - z_vector * np.dot(x0_vector[0, :], z_vector[0, :]) # convert to 1D vectors
        x_vector = x_vector / np.linalg.norm(x_vector) # vector 2 --> X
        x_vector2 = x0_vector2 - z_vector * np.dot(x0_vector2[0, :], z_vector[0, :])
        x_vector2 = x_vector2 / np.linalg.norm(x_vector2)
        y_vector = get_Y(z_vector, x_vector)
        y_vector2 = get_Y(z_vector, x_vector2)
        neighbors_transformed = transform(detection, x_vector, y_vector, z_vector, neighbors) # (N-1) x 3
        neighbors_transformed2 = transform(detection, x_vector2, y_vector2, z_vector, neighbors) # (N-1) x 3
        sc.append(get_shape_context(neighbors_transformed, mean_distance))
        sc2.append(get_shape_context(neighbors_transformed2, mean_distance))
    return np.array(sc), np.array(sc2)